@@ -27,16 +27,17 @@ import java.net.MalformedURLException;
 import java.net.URI;
 import java.net.URISyntaxException;
 import java.net.URL;
-import java.security.SecureRandom;
 import java.util.ArrayList;
 import java.util.Arrays;
 import java.util.Collection;
 import java.util.HashMap;
 import java.util.HashSet;
 import java.util.Iterator;
+import java.util.LinkedHashSet;
 import java.util.LinkedList;
 import java.util.List;
 import java.util.Map;
+import java.util.Random;
 import java.util.Set;
 import java.util.concurrent.atomic.AtomicBoolean;
 import java.util.jar.JarInputStream;
@@ -100,12 +101,14 @@ public class DirectoryWatcher extends Thread implements BundleListener
     public final static String NO_INITIAL_DELAY = "felix.fileinstall.noInitialDelay";
     public final static String DISABLE_CONFIG_SAVE = "felix.fileinstall.disableConfigSave";
     public final static String ENABLE_CONFIG_SAVE = "felix.fileinstall.enableConfigSave";
+    public final static String CONFIG_ENCODING = "felix.fileinstall.configEncoding";
     public final static String START_LEVEL = "felix.fileinstall.start.level";
     public final static String ACTIVE_LEVEL = "felix.fileinstall.active.level";
     public final static String UPDATE_WITH_LISTENERS = "felix.fileinstall.bundles.updateWithListeners";
     public final static String OPTIONAL_SCOPE = "felix.fileinstall.optionalImportRefreshScope";
     public final static String FRAGMENT_SCOPE = "felix.fileinstall.fragmentRefreshScope";
     public final static String DISABLE_NIO2 = "felix.fileinstall.disableNio2";
+    public final static String SUBDIR_MODE = "felix.fileinstall.subdir.mode";
 
     public final static String SCOPE_NONE = "none";
     public final static String SCOPE_MANAGED = "managed";
@@ -114,10 +117,6 @@ public class DirectoryWatcher extends Thread implements BundleListener
     public final static String LOG_STDOUT = "stdout";
     public final static String LOG_JUL = "jul";
 
-    static final SecureRandom random = new SecureRandom();
-
-    final File javaIoTmpdir = new File(System.getProperty("java.io.tmpdir"));
-
     final FileInstall fileInstall;
 
     Map<String, String> properties;
@@ -130,6 +129,7 @@ public class DirectoryWatcher extends Thread implements BundleListener
     boolean useStartActivationPolicy;
     String filter;
     BundleContext context;
+    private Bundle systemBundle;
     String originatingFileName;
     boolean noInitialDelay;
     int startLevel;
@@ -151,6 +151,9 @@ public class DirectoryWatcher extends Thread implements BundleListener
     // Represents installed artifacts which need to be started later because they failed to start
     Set<Bundle> delayedStart = new HashSet<Bundle>();
 
+    // Represents consistently failing bundles
+    Set<Bundle> consistentlyFailingBundles = new HashSet<Bundle>();
+
     // Represents artifacts that could not be installed
     final Map<File, Artifact> installationFailures = new HashMap<File, Artifact>();
 
@@ -164,6 +167,7 @@ public class DirectoryWatcher extends Thread implements BundleListener
         this.fileInstall = fileInstall;
         this.properties = properties;
         this.context = context;
+        systemBundle = context.getBundle(Constants.SYSTEM_BUNDLE_LOCATION);
         poll = getLong(properties, POLL, 2000);
         logLevel = getInt(properties, LOG_LEVEL, Util.getGlobalLogLevel(context));
         originatingFileName = properties.get(FILENAME);
@@ -185,12 +189,12 @@ public class DirectoryWatcher extends Thread implements BundleListener
         this.context.addBundleListener(this);
 
         if (disableNio2) {
-            scanner = new Scanner(watchedDirectory, filter);
+            scanner = new Scanner(watchedDirectory, filter, properties.get(SUBDIR_MODE));
         } else {
             try {
-                scanner = new WatcherScanner(context, watchedDirectory, filter);
+                scanner = new WatcherScanner(context, watchedDirectory, filter, properties.get(SUBDIR_MODE));
             } catch (Throwable t) {
-                scanner = new Scanner(watchedDirectory, filter);
+                scanner = new Scanner(watchedDirectory, filter, properties.get(SUBDIR_MODE));
             }
         }
     }
@@ -298,10 +302,10 @@ public class DirectoryWatcher extends Thread implements BundleListener
 
         while (!interrupted()) {
             try {
-                FrameworkStartLevel startLevelSvc = context.getBundle(0).adapt(FrameworkStartLevel.class);
+                FrameworkStartLevel startLevelSvc = systemBundle.adapt(FrameworkStartLevel.class);
                 // Don't access the disk when the framework is still in a startup phase.
                 if (startLevelSvc.getStartLevel() >= activeLevel
-                        && context.getBundle(0).getState() == Bundle.ACTIVE) {
+                        && systemBundle.getState() == Bundle.ACTIVE) {
                     Set<File> files = scanner.scan(false);
                     // Check that there is a result.  If not, this means that the directory can not be listed,
                     // so it's presumably not a valid directory (it may have been deleted by someone).
@@ -314,6 +318,7 @@ public class DirectoryWatcher extends Thread implements BundleListener
                     wait(poll);
                 }
             } catch (InterruptedException e) {
+                interrupt();
                 return;
             } catch (Throwable e) {
                 try {
@@ -332,7 +337,7 @@ public class DirectoryWatcher extends Thread implements BundleListener
         int type = bundleEvent.getType();
         if (type == BundleEvent.UNINSTALLED)
         {
-            for (Iterator it = getArtifacts().iterator(); it.hasNext();)
+            for (Iterator<?> it = getArtifacts().iterator(); it.hasNext();)
             {
                 Artifact artifact = (Artifact) it.next();
                 if (artifact.getBundleId() == bundleEvent.getBundle().getBundleId())
@@ -507,6 +512,8 @@ public class DirectoryWatcher extends Thread implements BundleListener
             delayedStart.removeAll(uninstalledBundles);
             // Try to start newly installed bundles, or bundles which we missed on a previous round
             startBundles(delayedStart);
+            consistentlyFailingBundles.clear();
+            consistentlyFailingBundles.addAll(delayedStart);
 
             // set the state as unchanged to not reattempt starting failed bundles
             setStateChanged(false);
@@ -593,17 +600,18 @@ public class DirectoryWatcher extends Thread implements BundleListener
     {
         if (tmpDir == null)
         {
+            File javaIoTmpdir = new File(System.getProperty("java.io.tmpdir"));
             if (!javaIoTmpdir.exists() && !javaIoTmpdir.mkdirs()) {
                 throw new IllegalStateException("Unable to create temporary directory " + javaIoTmpdir);
             }
-            for (;;)
+            Random random = new Random();
+            while (tmpDir == null)
             {
                 File f = new File(javaIoTmpdir, "fileinstall-" + Long.toString(random.nextLong()));
                 if (!f.exists() && f.mkdirs())
                 {
                     tmpDir = f;
                     tmpDir.deleteOnExit();
-                    break;
                 }
             }
         }
@@ -671,7 +679,7 @@ public class DirectoryWatcher extends Thread implements BundleListener
      */
     void refresh(Collection<Bundle> bundles) throws InterruptedException
     {
-        FileInstall.refresh(context, bundles);
+        FileInstall.refresh(systemBundle, bundles);
     }
 
     /**
@@ -941,14 +949,7 @@ public class DirectoryWatcher extends Thread implements BundleListener
                 URL transformed = artifact.getTransformedUrl();
                 String location = transformed.toString();
                 BufferedInputStream in = new BufferedInputStream(transformed.openStream());
-                try
-                {
-                    bundle = installOrUpdateBundle(location, in, artifact.getChecksum(), modified);
-                }
-                finally
-                {
-                    in.close();
-                }
+                bundle = installOrUpdateBundle(location, in, artifact.getChecksum(), modified);
                 artifact.setBundleId(bundle.getBundleId());
             }
             // if the listener is an artifact transformer
@@ -962,14 +963,7 @@ public class DirectoryWatcher extends Thread implements BundleListener
                 File transformed = artifact.getTransformed();
                 String location = path.toURI().normalize().toString();
                 BufferedInputStream in = new BufferedInputStream(new FileInputStream(transformed != null ? transformed : path));
-                try
-                {
-                    bundle = installOrUpdateBundle(location, in, artifact.getChecksum(), modified);
-                }
-                finally
-                {
-                    in.close();
-                }
+                bundle = installOrUpdateBundle(location, in, artifact.getChecksum(), modified);
                 artifact.setBundleId(bundle.getBundleId());
             }
             installationFailures.remove(path);
@@ -991,53 +985,62 @@ public class DirectoryWatcher extends Thread implements BundleListener
         String bundleLocation, BufferedInputStream is, long checksum, AtomicBoolean modified)
         throws IOException, BundleException
     {
-        is.mark(256 * 1024);
-        JarInputStream jar = new JarInputStream(is);
-        Manifest m = jar.getManifest();
-        if( m == null ) {
-            throw new BundleException(
-                "The bundle " + bundleLocation + " does not have a META-INF/MANIFEST.MF! "+
-                    "Make sure, META-INF and MANIFEST.MF are the first 2 entries in your JAR!");
-        }
-        String sn = m.getMainAttributes().getValue(Constants.BUNDLE_SYMBOLICNAME);
-        String vStr = m.getMainAttributes().getValue(Constants.BUNDLE_VERSION);
-        Version v = vStr == null ? Version.emptyVersion : Version.parseVersion(vStr);
-        Bundle[] bundles = context.getBundles();
-        for (Bundle b : bundles) {
-            if (b.getSymbolicName() != null && b.getSymbolicName().equals(sn)) {
-                vStr = b.getHeaders().get(Constants.BUNDLE_VERSION);
-                Version bv = vStr == null ? Version.emptyVersion : Version.parseVersion(vStr);
-                if (v.equals(bv)) {
-                    is.reset();
-                    if (Util.loadChecksum(b, context) != checksum) {
-                        log(Logger.LOG_WARNING,
-                                "A bundle with the same symbolic name ("
-                                        + sn + ") and version (" + vStr
-                                        + ") is already installed.  Updating this bundle instead.", null
-                        );
-                        stopTransient(b);
-                        Util.storeChecksum(b, checksum, context);
-                        b.update(is);
-                        modified.set(true);
+        JarInputStream jar = null;
+        try {
+            is.mark(256 * 1024);
+            jar = new JarInputStream(is);
+            Manifest m = jar.getManifest();
+            if( m == null ) {
+                throw new BundleException(
+                        "The bundle " + bundleLocation + " does not have a META-INF/MANIFEST.MF! " +
+                                "Make sure, META-INF and MANIFEST.MF are the first 2 entries in your JAR!");
+            }
+            String sn = m.getMainAttributes().getValue(Constants.BUNDLE_SYMBOLICNAME);
+            String vStr = m.getMainAttributes().getValue(Constants.BUNDLE_VERSION);
+            Version v = vStr == null ? Version.emptyVersion : Version.parseVersion(vStr);
+            Bundle[] bundles = context.getBundles();
+            for (Bundle b : bundles) {
+                if (b.getSymbolicName() != null && b.getSymbolicName().equals(sn)) {
+                    vStr = b.getHeaders().get(Constants.BUNDLE_VERSION);
+                    Version bv = vStr == null ? Version.emptyVersion : Version.parseVersion(vStr);
+                    if (v.equals(bv)) {
+                        is.reset();
+                        if (Util.loadChecksum(b, context) != checksum) {
+                            log(Logger.LOG_WARNING,
+                                    "A bundle with the same symbolic name ("
+                                            + sn + ") and version (" + vStr
+                                            + ") is already installed.  Updating this bundle instead.", null
+                            );
+                            stopTransient(b);
+                            Util.storeChecksum(b, checksum, context);
+                            b.update(is);
+                            modified.set(true);
+                        }
+                        return b;
                     }
-                    return b;
                 }
             }
-        }
-        is.reset();
-        Util.log(context, Logger.LOG_INFO, "Installing bundle " + sn
-                + " / " + v, null);
-        Bundle b = context.installBundle(bundleLocation, is);
-        Util.storeChecksum(b, checksum, context);
-        modified.set(true);
+            is.reset();
+            Util.log(context, Logger.LOG_INFO, "Installing bundle " + sn
+                    + " / " + v, null);
+            Bundle b = context.installBundle(bundleLocation, is);
+            Util.storeChecksum(b, checksum, context);
+            modified.set(true);
 
-        // Set default start level at install time, the user can override it if he wants
-        if (startLevel != 0)
+            // Set default start level at install time, the user can override it if he wants
+            if (startLevel != 0) {
+                b.adapt(BundleStartLevel.class).setStartLevel(startLevel);
+            }
+
+            return b;
+        }
+        finally
         {
-            b.adapt(BundleStartLevel.class).setStartLevel(startLevel);
+            if (jar != null)
+            {
+                jar.close();
+            }
         }
-        
-        return b;
     }
 
     /**
@@ -1190,8 +1193,8 @@ public class DirectoryWatcher extends Thread implements BundleListener
      */
     private void startAllBundles()
     {
-        FrameworkStartLevel startLevelSvc = context.getBundle(0).adapt(FrameworkStartLevel.class);
-        List<Bundle> bundles = new ArrayList<Bundle>();
+        FrameworkStartLevel startLevelSvc = systemBundle.adapt(FrameworkStartLevel.class);
+        Set<Bundle> bundles = new LinkedHashSet<>();
         for (Artifact artifact : getArtifacts()) {
             if (artifact.getBundleId() > 0) {
                 Bundle bundle = context.getBundle(artifact.getBundleId());
@@ -1210,11 +1213,13 @@ public class DirectoryWatcher extends Thread implements BundleListener
      /**
       * Starts a bundle and removes it from the Collection when successfully started.
       */
-    private void startBundles(Collection<Bundle> bundles)
+    private void startBundles(Set<Bundle> bundles)
     {
+        // Check if this is the consistent set of bundles which failed previously.
+        boolean logFailures = !consistentlyFailingBundles.equals(bundles);
         for (Iterator<Bundle> b = bundles.iterator(); b.hasNext(); )
         {
-            if (startBundle(b.next()))
+            if (startBundle(b.next(), logFailures))
             {
                 b.remove();
             }
@@ -1226,9 +1231,9 @@ public class DirectoryWatcher extends Thread implements BundleListener
       * @param bundle the bundle to start.
       * @return whether the bundle was started.
       */
-    private boolean startBundle(Bundle bundle)
+    private boolean startBundle(Bundle bundle, boolean logFailures)
     {
-        FrameworkStartLevel startLevelSvc = context.getBundle(0).adapt(FrameworkStartLevel.class);
+        FrameworkStartLevel startLevelSvc = systemBundle.adapt(FrameworkStartLevel.class);
         // Fragments can never be started.
         // Bundles can only be started transient when the start level of the framework is high
         // enough. Persistent (i.e. non-transient) starts will simply make the framework start the
@@ -1249,7 +1254,10 @@ public class DirectoryWatcher extends Thread implements BundleListener
             catch (BundleException e)
             {
                 // Don't log this as an error, instead we start the bundle repeatedly.
-                log(Logger.LOG_WARNING, "Error while starting bundle: " + bundle.getLocation(), e);
+                if (logFailures)
+                {
+                    log(Logger.LOG_WARNING, "Error while starting bundle: " + bundle.getLocation(), e);
+                }
             }
         }
         return false;
