@@ -19,11 +19,15 @@
 package org.apache.felix.fileinstall.internal;
 
 import java.io.File;
+import java.io.FileOutputStream;
+import java.io.OutputStream;
+import java.nio.file.Files;
 import java.util.Dictionary;
 import java.util.Hashtable;
 import java.util.concurrent.atomic.AtomicReference;
 
 import junit.framework.TestCase;
+import org.easymock.Capture;
 import org.easymock.EasyMock;
 import org.easymock.IAnswer;
 import org.easymock.IArgumentMatcher;
@@ -71,11 +75,73 @@ public class ConfigInstallerTest extends TestCase {
         assertEquals( "Pid with Factory Pid calculated", "pid", ci.parsePid( path )[1] );
     }
 
+    public void testTypedConfiguration() throws Exception
+    {
+        File file = File.createTempFile("test", ".config");
+        try (OutputStream os = new FileOutputStream(file)) {
+            os.write("networkInterface=\"wlp3s0\"\n".getBytes("UTF-8"));
+        }
+        String pid = file.getName().substring(0, file.getName().indexOf(".config"));
+
+        Capture<Dictionary<String, Object>> props = new Capture<>();
+        EasyMock.expect(mockConfigurationAdmin.listConfigurations((String) EasyMock.anyObject()))
+                .andReturn(null);
+        EasyMock.expect(mockConfigurationAdmin.getConfiguration(pid, "?"))
+                .andReturn(mockConfiguration);
+        EasyMock.expect(mockConfiguration.getProperties())
+                .andReturn(null);
+        EasyMock.expect(mockBundleContext.getProperty((String) EasyMock.anyObject()))
+                .andReturn(null)
+                .anyTimes();
+        mockConfiguration.update(EasyMock.capture(props));
+        EasyMock.expectLastCall();
+        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
+
+        ConfigInstaller ci = new ConfigInstaller( mockBundleContext, mockConfigurationAdmin, new FileInstall() );
+        ci.install(file);
+
+        EasyMock.verify(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
+        Dictionary<String, Object> loaded = props.getValue();
+        assertNotNull(loaded);
+        assertEquals("wlp3s0", loaded.get("networkInterface"));
+    }
+
+    public void testTypedConfigurationFloat() throws Exception
+    {
+        File file = File.createTempFile("test", ".config");
+        try (OutputStream os = new FileOutputStream(file)) {
+            os.write("key=F\"1137191584\"\n".getBytes("UTF-8"));
+        }
+        String pid = file.getName().substring(0, file.getName().indexOf(".config"));
+
+        Capture<Dictionary<String, Object>> props = new Capture<>();
+        EasyMock.expect(mockConfigurationAdmin.listConfigurations((String) EasyMock.anyObject()))
+                .andReturn(null);
+        EasyMock.expect(mockConfigurationAdmin.getConfiguration(pid, "?"))
+                .andReturn(mockConfiguration);
+        EasyMock.expect(mockConfiguration.getProperties())
+                .andReturn(null);
+        EasyMock.expect(mockBundleContext.getProperty((String) EasyMock.anyObject()))
+                .andReturn(null)
+                .anyTimes();
+        mockConfiguration.update(EasyMock.capture(props));
+        EasyMock.expectLastCall();
+        EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
+
+        ConfigInstaller ci = new ConfigInstaller( mockBundleContext, mockConfigurationAdmin, new FileInstall() );
+        ci.install(file);
+
+        EasyMock.verify(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
+        Dictionary<String, Object> loaded = props.getValue();
+        assertNotNull(loaded);
+        assertEquals(400.333F, loaded.get("key"));
+    }
+
     public void testGetNewFactoryConfiguration() throws Exception
     {
         EasyMock.expect(mockConfigurationAdmin.listConfigurations((String) EasyMock.anyObject()))
                     .andReturn(null);
-        EasyMock.expect(mockConfigurationAdmin.createFactoryConfiguration( "pid", null ))
+        EasyMock.expect(mockConfigurationAdmin.createFactoryConfiguration( "pid", "?" ))
                     .andReturn(mockConfiguration);
         EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
 
@@ -91,7 +157,7 @@ public class ConfigInstallerTest extends TestCase {
     {
         EasyMock.expect(mockConfigurationAdmin.listConfigurations((String) EasyMock.anyObject()))
                         .andReturn(null);
-        EasyMock.expect(mockConfigurationAdmin.createFactoryConfiguration( "pid", null ))
+        EasyMock.expect(mockConfigurationAdmin.createFactoryConfiguration( "pid", "?" ))
                         .andReturn(mockConfiguration);
         EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
 
@@ -107,7 +173,7 @@ public class ConfigInstallerTest extends TestCase {
     {
         EasyMock.expect(mockConfigurationAdmin.listConfigurations((String) EasyMock.anyObject()))
                         .andReturn(null);
-        EasyMock.expect(mockConfigurationAdmin.getConfiguration( "pid", null ))
+        EasyMock.expect(mockConfigurationAdmin.getConfiguration( "pid", "?" ))
                         .andReturn(mockConfiguration);
         EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
 
@@ -126,7 +192,7 @@ public class ConfigInstallerTest extends TestCase {
         EasyMock.expect(mockBundleContext.getProperty(DirectoryWatcher.LOG_LEVEL)).andReturn(null);
         EasyMock.expect(mockConfigurationAdmin.listConfigurations((String) EasyMock.anyObject()))
                         .andReturn(null);
-        EasyMock.expect(mockConfigurationAdmin.getConfiguration("pid", null ))
+        EasyMock.expect(mockConfigurationAdmin.getConfiguration("pid", "?" ))
                         .andReturn(mockConfiguration);
         EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
 
@@ -140,6 +206,7 @@ public class ConfigInstallerTest extends TestCase {
 
     public void testSetConfiguration() throws Exception
     {
+        EasyMock.expect(mockBundleContext.getProperty(DirectoryWatcher.CONFIG_ENCODING)).andReturn(null);
         EasyMock.expect(mockBundleContext.getProperty(DirectoryWatcher.LOG_DEFAULT)).andReturn(null);
         EasyMock.expect(mockBundleContext.getProperty(DirectoryWatcher.LOG_LEVEL)).andReturn(null);
         EasyMock.expect(mockConfiguration.getProperties()).andReturn(new Hashtable<String, Object>());
@@ -158,7 +225,7 @@ public class ConfigInstallerTest extends TestCase {
         mockConfiguration.update(new Hashtable<String, Object>());
         EasyMock.expect(mockConfigurationAdmin.listConfigurations((String) EasyMock.anyObject()))
                         .andReturn(null);
-        EasyMock.expect(mockConfigurationAdmin.getConfiguration("firstcfg", null))
+        EasyMock.expect(mockConfigurationAdmin.getConfiguration("firstcfg", "?"))
                         .andReturn(mockConfiguration);
         EasyMock.replay(mockConfiguration, mockConfigurationAdmin, mockBundleContext);
 
